/*
 ini.c

 Created: 2013.03.25
*/

#include <enviro64.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "interpreter.h"
#include "ini.h"

/* ---- code */

int deltobj(tobj *o)
{
	if (o->z>0 && o->s)
		{free(o->s); o->s = NULL, o->z = 0; return 1;}
	return 0;
}

robj *getrobj(const tcon *l,
	const tobj *o)
{
	long		z = o->z&0x7FFFFFFF;

	for(; l->d; ++l)
		if (!memcmp(o->s,l->s,z) && !l->s[z])
			return l->d;
	return NULL;
}

//#define GET_robj_DATA_SIZE	((z)<-255 ? 1u<<(-(z)&3):ABS(z))!=0)

int setrobj(robj *l,
	const tobj *d,
	const tobj *s)
{
	int8		*p;
	long		z = d->z&0x7FFFFFFF, t;

	while(1)
	{
		if ((t=ABS(l->z))==0)
			return -1;
		if (!memcmp(d->s,l->s,z) && !l->s[z])
			break;
		++l;
	}
	p = l->p, z = s->z&0x7FFFFFFF;
	if (l->z<-255)
	{
		double	v;

		if (!parse_number(s->s,s->s+z,&v,1))
			if (z==4 && !memcmp(s->s,"true",4))
				v = 1;
			else if (z==5 && !memcmp(s->s,"false",4))
				v = 0;
			else
				return -1;
		switch(l->z)
		{
		case _int8: if (v>IMAX(int8)) *(int8*)p = IMAX(int8); else *(int8*)p = v; break;
		case _int16: if (v>IMAX(int16)) *(int16*)p = IMAX(int16); else *(int16*)p = v; break;
		case _ifract: v*= 16384;
		case _ifixed: v*= 65536, v+= v<0 ? -.5:.5;
		case _int32: if (v>IMAX(int32)) *(int32*)p = IMAX(int32); else *(int32*)p = v; break;
		case _int64: if (v>=-IMIN64) int64sethl(*(int64*)p,IMAX(int32),-1ul); else int64xset(*(int64*)p,v); break;
		case _uint8: if (v>0xFF) *(uint8*)p = 0xFF; else *(uint8*)p = v; break;
		case _ufrc16: v*= 65536, v+= v<0 ? -.5:.5;
		case _uint16: if (v>0xFFFF) *(uint16*)p = 0xFFFF; else *(uint16*)p = v; break;
		case _ufract: v*= 16384;
		case _ufixed: v*= 65536, v+= v<0 ? -.5:.5;
		case _uint32: if (v>-1ul) *(uint32*)p = -1ul; else *(uint32*)p = v; break;
		case _uint64: if (v>=-2*IMIN64) int64sethl(*(int64*)p,-1ul,-1ul); else int64xset(*(uint64*)p,v); break;
		case -264: *(int8*)p = v!=0; break;
//		case -269: *(int16*)p = ((int16*)&v)[!BIG_ENDIAN*3]; break;
		case _float: *(float*)p = v; break;
		case _double: *(double*)p = v; break;

		default: return -1;
		}
	} else
	{
		if (l->z<0)
			++p;
		if (z>=t)
			z = t;
		else
			p[z] = 0; /* C termination */
		if (l->z<0)
			p[-1] = z;
		memcpy(p,s->s,z);
	}
	return 0;
}

static int parse_line(const char *str,
	const char *e,
	const tcon *l,
	robj **c)
{
	tobj		o[2];
	const char	*s;
	int		t = 0;

	o->s = o[1].s = NULL;
	if ((s=str)==NULL)
		return -1;
	while(s<e)
	{
		switch((uint8)*s) /* try not to get stuck */
		{
		case ' ':
			s = parse_skipwsp(s,e);
			break;
		case '=':
			if (t)
				goto err;
			++t, ++s;
			break;
		case '.':
			if (!t)
				goto err;
		case '-':
		case '+':
		CASE_NUMERIC
		CASE_ALPHABETIC
			if (t>1 || o[t].s)
				goto err;
			o[t].s = (char*)s; /* shouldn't be altered */
			while((s=parse_wordbreak(s,e,NULL))<e)
				if (*s=='-' || *s=='+' || *s=='e' || *s=='E')
					++s;
				else if (*s=='.')
				{ /* only one dot in a row */
					if (++s<e && *s=='.')
						goto err;
				} else
					break;
			o[t].z = (s-o[t].s)|0x80000000u;
			if (!t)
				break;
			goto obj;
		case '"':
		case '\047':
			if (!t || o[t].s)
				goto err;
			if ((s=parse_textloop(s,e,&o[t].z,NULL,&o[t].s,0,NULL))==NULL
					|| o[t].z<0)
				{o[t].z = 1; goto err;}
			++s;
		obj:	if (setrobj(*c,o,o+1))
				goto err;
			break;
		case '[':
			if (t)
				goto err;
			s = parse_skipwsp(s+1,e);
			if (s>=e)
				goto err;
			s = parse_wordbreak(o->s=(char*)s,e,NULL);
			o->z = (s-o->s)|0x80000000u;
			s = parse_skipwsp(s,e);
			if (s>=e || *s++!=']')
				goto err;
			if ((*c=getrobj(l,o))==NULL)
				goto err;
			break;

		default:
		err:	t = s-str;
			goto end;
		}
	}
	t = 0;
end:	deltobj(o), deltobj(o+1);
	return t;
}

long read_ini(FILE *f,
	const tcon *cl)
{
	robj		*c = NULL;
	char		*s;
	uint32		z, l, x = 1;

	s = malloc(0); /* needed for first-empty line(s) case */
	while(z=~0,l=x,f && (s=freadline(f,&z,s,NULL,&x,'\\'))!=NULL)
	{
		const char	*e = s+z;
		long		r;

		if ((r=parse_line(parse_skipwsp(s,e),e,cl,&c))!=0)
			{z = l; goto end;}
		if (!f)
			{z = 0; break;}
	}
	if (z)
		z = _e_memory_error;
end:	free(s);

	return z;
}

static int fixfrac(char *b,
	uint16 v)
{
	char		*s;
	uint32		m, f;
	int		i;

	if ((m=(uint32)v*6250+3125>>13)==0)
		{*b = 0; return 0;}
	m+= m, *(s=b+6) = 0, v = 0, i = 5;
	do
	{
		f = m/10;
		if ((*--s=m-f*10)!=0 || v)
			*s+= '0', ++v;
		m = f;
	} while(--i);
	*--s = '.';
	return v+1;
}

int print_robj(FILE *f,
	const robj *l)
{
	const char	*p = l->p;
	long		z, r;

	if (l->z<-255)
	{
		char		b[8];
		double		v;

		switch(l->z)
		{
		case _int8: z = *(int8*)p; goto i10;
		case _int16: z = *(int16*)p; goto i10;
		case _int32: z = *(int32*)p;
		i10:	if ((r=z)<0)
				z =-z;
			goto u10;
		case _int64:
			return fprintf(f,"%s=%lld",
				l->s,*(int64*)p);
		case _uint8: z = *(uint8*)p; goto z10;
		case _uint16: z = *(uint16*)p; goto z10;
		case _uint32: z = *(uint32*)p;
		z10:	r = 0;
		u10:	*b = 0;
			goto z16;
		case _uint64:
			return fprintf(f,"%s=%llu",
				l->s,*(uint64*)p);
		case _bool: z = *(int8*)p!=0; goto z10;
		case _ufrc16:
			if ((z=*(uint16*)p)==0)
				goto z10;
			fixfrac(b,z);
			return fprintf(f,"%s=%s",
				l->s,b);
		case _ifixed: z = r = *(int32*)p, z = ABS(z); goto x16;
		case _ufixed: z = *(uint32*)p, r = 0;
		x16:	if (!z)
				goto u10;
			fixfrac(b,z);
			z>>= 16;
		z16:	return fprintf(f,r<0 ? "%s=-%u%s":"%s=%u%s",
				l->s,z,b);
		case _ifract: v = *(int32*)p/1073741824.0; break;
		case _ufract: v = *(uint32*)p/1073741824.0; break;
		case _float: v = *(float*)p;
			return fprintf(f,"%s=%.8g",
				l->s,v);
		case _double: v = *(double*)p;
			return fprintf(f,"%s=%.16g",
				l->s,v);

		default: return -1;
		}
		return fprintf(f,"%s=%.10g",
			l->s,v);
	} else
	{
		r = fprintf(f,"%s=\"",
			l->s);
		if ((z=l->z)<0)
		{
			z =-z;
			if (*(uint8*)p<z)
				z = *(uint8*)p;
			++p;
		}
		r+= print_chars(f,p,z,'"');
		r+= fputc('"',f);
		return r;
	}
}

void dump_ini(FILE *f,
	const tcon *l)
{
	const robj	*r;

	for(; (r=l->d)!=0; ++l)
		for(fprintf(f,"[%s]\n", l->s); r->z; ++r)
			print_robj(f,r), fputc('\n',f);
}

#ifdef EXAMPLE
struct graphics_data
{
	double		bitdepth;
	char		bullscrap[8];
} graphics;

static robj graphics_context[] =
{
	{_ifixed, "bitdepth", &graphics.bitdepth},
	{-7, "bullisht", &graphics.bullscrap},
	{0}
};

static tcon context_list[] =
{
	{graphics_context, "graphics"},
	{NULL},
};

int main(int argc,
	char **argv)
{
	FILE		*f = NULL;
	char		**a, *s;
	int		r = 0;

	if (argc<=0 || (a=argv)==NULL)
		return 0;
	if (!--argc || (s=*++a)==NULL)
	{
		fprintf(stderr,"### usage: %s filename\n",
			*argv);
		fputs(
"#\n"
"#  -            use stdin\n"
"#  -E<expr>     evaluate expression\n"
"#\n",
			stderr);
		return 1;
	}

	if (*s=='-')
	{
		if (!s[1])
			f = stdin, s = "dev::stdin";
		else
		{
			fprintf(stderr,"### %s - unknown option '%s'.\n",
				*argv,s);
			return 1;
		}
	} else
		f = fopen(s,"r");
	if (!f)
	{
		fprintf(stderr,"### %s - could not get/open '%s'.\n",
			*argv,s);
		return 1;
	}
#if !TARGET_API_MAC_CARBON && TARGET_OS_MAC
	if (((volatile char*)0x174)[60>>3]&(0x80>>(60&7))) /* keymap - control */
		Debugger();
#endif
	r = read_ini(f,context_list);
	if (f!=stdin)
		fclose(f);
	if (r)
	{
		if (r<0)
			fprintf(stderr,"### %s - fatal error (%d).\n",
				*argv,r);
		else if (f)
			fprintf(stderr,
"File \"%s\"; line %d\n",
				s,r);
		r = 1;
	} else
		dump_ini(stdout,context_list);

	return r;
}
#endif
