/*
 lists.c

  Created: 2011.09.26 23:09 EDT
*/

#include <stdlib.h>
#include "interpreter.h"

#if __POCC__
 #include <string.h>
 /* generates native code */
 #define _rcopy				memcpy
#else
 #define _rcopy(d,s,z)			do{uint8 *__d=d; const uint8 *__s=s; \
	if(z>3) {*__d++=*__s++,*__d++=*__s++,*__d++=*__s++,*__d++=*__s++; \
		if(z>7) {*__d++=*__s++,*__d++=*__s++,*__d++=*__s++,*__d=*__s; \
			break; \
		} \
	} \
	if(z&2) *__d++=*__s++,*__d++=*__s++; \
	if(z&1) *__d=*__s; \
}while(0)
#endif

list_item *list_add(list_t *p,
	int32 x,
	int32 z,
	void *v)
{
	list_item	*l;
	int32		c;

	if (!p)
		{l = NULL; goto blank;}
	if ((l=*p)==NULL || (c=l->index)<0)
		{blank: c = 0;}
	if ((l=realloc(l,(2u+c)<<4))==NULL)
		return NULL;
	if (p)
		*p = l;
	if (!c)
		l->type = 0, l->data.ptr = 0;
	l+= (unsigned)(l->index=1+c), l->index = x;
	if ((l->type=z)<=0 || ITEM_HAS_MALLOC(l))
		l->data.ptr = v;
	else
	{
		l->data.ptr = 0;
		if (ITEM_IS_RAWVAL(l))
			z = RAW_VAL_SIZE(z);
		_rcopy(l->data.raw,v,z);
	}
	return l; //c ? l-1:l
}

void list_dispose(list_t p)
{
	list_item	*l;
	int32		c;

	if ((l=p)==NULL)
		return;
	if ((c=l->index)>0) /* bit #31 not allowed in list */
		do
			if (l->data.lptr!=p)
				if (ITEM_HAS_MALLOC(l))
					{free(l->data.ptr); l->data.ptr = 0;}
				else if (l->type==_t_list) /* no refs */
					list_dispose(l->data.lptr); /* be careful */
		while(++l,--c);
	free(p);
}

/* ---- items */

/* omit stored references (external variables) in writing list to file */

int32 item_getsize(const list_item *l)
{
	int32		t;

	if ((t=l->type&-1u>>1)>0x7FFFFFF0)
		if (t==_t_list)
			return (unsigned)l->data.lptr->index+1<<4;
		else
			return RAW_VAL_SIZE(t);
	return t;
}

/* adds to current size; returns last position (end if subtracting) */
void *item_resize(list_item *l,
	int32 z)
{
	void		*p;
	int32		t, c = 0;

	if (!l)
		return NULL;
	if ((t=l->type&-1u>>1)>0x7FFFFFF0)
	{
		if (ITEM_IS_REF(l))
			return NULL;//-3; /* reference only modified by owner */
		if (t==_t_list) /* be sure to dispose of mallocs first */
			t = (unsigned)l->data.lptr->index+1<<4, c = 1;
		else
			t = RAW_VAL_SIZE(t);
	}
	if ((t+=z)==0)
	{ /* now empty; deallocate ptr */
		if (ITEM_HAS_MALLOC(l))// && l->data)
			l->data.ptr = realloc(l->data.ptr,0);
		else if (ITEM_IS_MALLOC_REF(l))
		{
			list_t	d;

			if ((d=l->data.lptr)==NULL || d==l)
				return NULL;
			if (d->index>0 && !ITEM_IS_REF(l))
				{list_dispose(d); l->data.lptr = NULL;}
		} else if ((l->type&-1u>>1)==_t_list)
		{
			if (l->data.lptr!=l)
				list_dispose(l->data.lptr);
			l->data.lptr = NULL;
		}
		l->type = 0;
		return l->data.raw;
	} else if ((unsigned)t>0x7FFFFFF0) /* out of range */
		return NULL;//-2;
	if (!((z=-z)&0x80000000U))
		z = 0; /* at end */
	z+= t;
	if (t>8)
	{
		if (ITEM_HAS_MALLOC(l))
		{
			if ((p=realloc(l->data.ptr,t))==NULL)
				return NULL;//-2;
		} else if ((p=realloc(NULL,t))==NULL)
			return NULL;//-2;
		else /* expanded beyond raw bounds */
			{_rcopy(p,ITEM_IS_REF(l) ? l->data.ptr:l->data.raw,z);}
		l->data.ptr = p, p = z+(int8*)p;
	} else
	{
		if (ITEM_HAS_MALLOC(l))
		{ /* convert to raw */
			p = l->data.ptr;
			_rcopy(l->data.raw,p,z);
			realloc(p,0);
		} else if (ITEM_IS_REF(l))
			_rcopy(l->data.raw,l->data.ptr,z); /* raw copy */
		p = l->data.raw+z;
	}
	/* should only update size until after buffered */
	if (c)
		l->data.lptr->index = (t-16)>>4; /* clear bit 31? */
	else
		l->type = t;
	return p;
}

int32 item_set(list_item *l,
	int32 x,
	int32 z,
	void *p,
	int r)
{
	int32		t;
	uint8		*d = l->data.raw;

	if (ITEM_IS_REF(l))
	{
		double	v;

		t = l->type&-1u>>1;
		if (ITEM_IS_MALLOC_REF(l))
		{
			list_item	*d;

			if ((d=l->data.lptr)==NULL || l==d)
				return 0;
			if (l->type==(_t_list|-1<<31))
				goto xfer; /* overwrite ref to list */
			//if (ITEM_IS_BIN(d))
			//	return 0; /* constant - should never get here */
			if ((unsigned)z<_t_list)
			{ /* reference is to variable-length string */
				if (x<0)
					x+= t;
				d->type = x|-1<<31;
				if (!item_resize(d,x=z-t))
					return -1;
				return z;
			}
		}
		/* store (with possible conversion) */
		if (t==z || (z&t&8 && (z&~3)==(t&~3))) /* match */
			goto copy;
		d = l->data.ptr;
	sraw:	v = item_to_double(z,p);
		switch(t) /* POCC produces bogus #2130 warning here */
		{
		case _t_float: *(float*)d = v; break;
		case _t_double: *(double*)d = v; break;
		case _t_ifixed: *(ifixed*)d = v*65536.0; break;
		case _t_ufixed: *(ufixed*)d = v*65536.0; break;
		case _t_int8: *(int8*)d = v; break;
		case _t_int16: *(int16*)d = v; break;
		case _t_int32: *(int32*)d = v; break;
		case _t_int64: int64xset(*(int64*)d,v);
		case _t_uint8: *(uint8*)d = v; break;
		case _t_uint16: *(uint16*)d = v; break;
		case _t_uint32: *(uint32*)d = v; break;
		case _t_uint64: int64xset(*(uint64*)d,v);
		}
		return RAW_VAL_SIZE(z);
	} else if (l->type==_t_list)
	{
		if (l!=l->data.lptr)
			list_dispose(l->data.lptr);
	} else if (ITEM_HAS_MALLOC(l))
		free(l->data.ptr);
	else if (r && ITEM_IS_RAWVAL(l))
		{t = l->type&-1u>>1; goto sraw;}
xfer:	l->index = x, l->type = z;
	if (ITEM_HAS_MALLOC(l) || ITEM_IS_REF(l)) /* blind transfer */
		{l->data.ptr = p; return z&-1u>>1;}
	if (ITEM_IS_RAWVAL(l))
	copy:	z = RAW_VAL_SIZE(z);
	_rcopy(d,p,z);

	return z;
}

double item_to_double(int32 z,
	const void *p)
{
	double		v = 0;

	if (!p)
		zero: return 0;
	if (z<=(_t_list|-1<<31))
	{
		const list_item	*s = p;
		int32		m = s->type&-1u>>1;
		/* note that str ref to ref str is a raw ptr not an item */
		/* the reference would otherwise be copied as is */
		if (ITEM_HAS_MALLOC(s) || ITEM_IS_REF(s))
		{
			if ((p=s->data.ptr)==0 || s==(list_item*)p)
				goto zero;
		} else
			p = s->data.raw;
		z&=-1u>>1;
		if ((z=m-z)<0)
			z = 0;
		else
			p = z+(int8*)p;
	} else
		z&=-1u>>1;
	switch(z)
	{
	case _t_list:
	case _t_func:
	case _t_binary:
		break;
	case _t_float: return *(const float*)p;
	case _t_double: return *(const double*)p;
	case _t_ifixed: return *(const ifixed*)p*(1.0/65536);
	case _t_ufixed: return *(const ufixed*)p*(1.0/65536);
	case _t_int8: return *(const int8*)p;
	case _t_int16: return *(const int16*)p;
	case _t_int32: return *(const int32*)p;
	case _t_int64: return int64tox(*(const int64*)p);
	case _t_uint8: return *(const uint8*)p;
	case _t_uint16: return *(const uint16*)p;
	case _t_uint32: return *(const uint32*)p;
	case _t_uint64: return int64tox(*(const uint64*)p);
	}
	{
		const char	*n = parse_number(p,z+(const char*)p,&v,1);

		if (n-(const char*)p!=z)
			*(int32*)&v =-1, ((int32*)&v)[1] = n ? n-(const char*)p:-1;
	}
	return v;
}

/* returns one-based index (to skip header item) or zero */
uint32 getnamesitemindex(const uint8 *s,
	int32 t,
	const char *n,
	uint8 z,
	uint8 h)
{
	const uint8	*e;
	uint32		i;
	uint8		c;

	if (t<=0)
		return 0;
	e = s+t, i = 1;
	do
		if (c=*s++,t=*s++,c==h && t==z && !strncasecmp(n,(const char*)s,t))
			return i;
	while(s+=t,++i,s<e);

	return 0;
}
