#if 0
C -o script lists.o script.o interpreter.o -s Interpreter=Main -di interpreter.h 
	-i :: -d mac -d EVAL_FLOATING -d ALLOW_RUNON_WSP {"Parameters"}
Exit 0
#endif
/*
 script.c
  basic interpreter (no functions or loops yet)

  Created: 2011.08.20
  Version: 1.6.4
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "interpreter.h"

/* ---- mid-level */

#if TARGET_OS_MAC
 #define commandPeriod()		(*((volatile char*)0x179)&*((volatile char*)0x17A))
#else
 #define commandPeriod()		0 /* you loose */
#endif

static uint8	user_abort, modal;

void global_idle_proc(void)
{
	if (!commandPeriod() && !user_abort)
		return;
	if (user_abort && !modal)
	{
		fputs("\n# User abort.\n",stderr);
		exit(1);
	}
	user_abort = 1;
}

/* ---- evaluation */

static evalvalue evalindir(evalvalue v)
{
	return ((unsigned)v&0xFFFF)*12345;
}

static const char *getvalue(const char *w,
	const char *e,
	evalvalue *i,
	void *u)
{
	list_item	*l;
	long		x;
	uint8		h;

	if ((e=parse_wordbreak(w,e,&h))==NULL)
		return NULL;
	if ((l=*(list_t*)u)!=NULL && (x=getnamesitemindex(ITEM_HAS_MALLOC(l) ?
			l->data.rptr:l->data.raw,l->type&-1U>>1,w+1,e-w-1,h))!=0)
	{
		l+= x;
		if (ITEM_IS_REF(l))
			*i = *l->data.vptr;
		else
			*i = l->data.val;
		return e;
	}
	return 0;
}

static const char *setvalue(const char *w,
	const char *e,
	evalvalue i,
	void *u)
{
	list_t		*p = (list_t*)u;
	list_item	*l;
	long		x;
	int		z;
	uint8		h;

	if ((e=parse_wordbreak(w,e,&h))==NULL || (z=e-w-1)>255)
		return NULL;
	if ((l=*p)!=NULL && (x=getnamesitemindex(ITEM_IS_REF(l) ? l->data.rptr
			:l->data.raw,l->type&-1U>>1,w+1,z,h))!=0)
		{
			if ((l+=x)->index&1U<<31) /* ID is binary */
				return NULL;
			else
				item_set(l,0,_t_eval,&i,0);
			goto end;
		}
	if ((l=list_add(p,0,_t_eval,&i))==NULL)
		return NULL;
	/* add name to list */
	if ((u=item_resize(*p,z+2))!=0)
	{
		*(char*)u = h, ((char*)u)[1] = z;
		memcpy(2+(char*)u,w+1,z);
	}
end:	return e;
}

/* ---- script */

char *err_strs[] =
{
	"syntax error",
	"memory error",
	"read error",
	"write error",
	"incomplete expression",
	"interrogator already open",
	"no such identifier",
	"handler not supplied",
	"stack over/underflow"
};

static long parse_script(FILE *f,
	char *s)
{
	list_t		thread = NULL;
	evalvalue	v = 0;
	uint32		z, l, x = 1;
	int		err = 0;

	setvalue("pi",(char*)-1,3.14159265358979323846,&thread); /* circumf./diameter */
	/* variable(s) initialized to value of their recognized constant */
	setvalue("C",(char*)-2,299792458,&thread); /* light in vaccuum, m/sec */
	if (f)
		s = malloc(0); /* needed for first-empty line(s) case */
	else
		{z = strlen(s), l = 1; goto eval;}
	while(z=~0,l=x,f && (s=freadline(f,&z,s,NULL,&x,'\\'))!=NULL)
	{
		const char	*e;
		long		r;
	eval:	/* suggestion: preprocess line with macros, including __LINE__ */
		e = s+z;
		r = evaluate_line(parse_skipwsp(s,e),e,&v,&err,getvalue,setvalue,&thread,evalindir);
		if (err)
		{
			if (err==_e_param_separation)
				err = _e_general_error;
			if (err<0 && (err=~err)<GET_ARRAY_COUNT(err_strs))
				fprintf(stderr,
"### %s\n"
"### % *^\n"
"### %s\n",
					s,r,err_strs[err]);
			z = l;
			goto end;
		}
		if (r==(-1u>>1) || !f)
			{z = 0; break;}
	}
	if (z)
		z = _e_memory_error;
	else
		printf("# result: %.16g\n",
			(double_t)v);
end:	list_dispose(thread);
	if (f)
		free(s);

	return z;
}

int main(int argc,
	char **argv)
{
	FILE		*f = NULL;
	char		**a, *s, *str = NULL;
	int		r = 0;

	if (argc<=0 || (a=argv)==NULL)
		return 0;
	if (!--argc || (s=*++a)==NULL)
	{
		fprintf(stderr,"### usage: %s filename\n",
			*argv);
		fputs(
"#\n"
"#  -            use stdin\n"
"#  -E<expr>     evaluate expression\n"
"#\n",
			stderr);
		return 1;
	}

	if (*s=='-')
	{
		if (!s[1])
			f = stdin, s = "dev::stdin";
		else if (s[1]=='E')
			str = s+2;
		else
		{
			fprintf(stderr,"### %s - unknown option '%s'.\n",
				*argv,s);
			return 1;
		}
	} else
		f = fopen(s,"r");
	if (!f && !str)
	{
		fprintf(stderr,"### %s - could not get/open '%s'.\n",
			*argv,s);
		return 1;
	}
#if !TARGET_API_MAC_CARBON && TARGET_OS_MAC
	if (((volatile char*)0x174)[60>>3]&(0x80>>(60&7))) /* keymap - control */
		Debugger();
#endif
	r = parse_script(f,str);
	if (f!=stdin)
		fclose(f);
	if (r)
	{
		if (r<0)
			fprintf(stderr,"### %s - fatal error (%d).\n",
				*argv,r);
		else if (f)
			fprintf(stderr,
"File \"%s\"; line %d\n",
				s,r);
		r = 1;
	}

	return r;
}
